#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from .amber.parser import AmberParser
from .asap.parser import AsapParser
from .dftbplus.parser import DFTBPlusParser
from .dlpoly.parser import DLPolyParser
from .gromacs.parser import GromacsParser
from .gromos.parser import GromosParser
from .gulp.parser import GulpParser
from .lammps.parser import LammpsParser
from .libatoms.parser import LibAtomsParser
from .mopac.parser import MopacParser
from .namd.parser import NAMDParser
from .openkim.parser import OpenKIMParser
from .tinker.parser import TinkerParser
